# -*- coding: utf-8 -*-

"""Handy utilities for creating curves."""

from math import pi, cos, sin, sqrt, ceil, atan2
import copy
import inspect

import numpy as np
from numpy.linalg import norm
import scipy.sparse as sp
import scipy.sparse.linalg as splinalg

from .curve import Curve
from .basis import BSplineBasis
from .utils import flip_and_move_plane_geometry, rotate_local_x_axis
from . import state

__all__ = ['Boundary', 'line', 'polygon', 'n_gon', 'circle', 'ellipse',
           'circle_segment_from_three_points', 'circle_segment', 'interpolate',
           'least_square_fit', 'cubic_curve', 'bezier', 'manipulate', 'fit']

class Boundary:
    """Enumeration representing different boundary conditions used in
    :func:`interpolate`."""

    FREE = 1
    """The curve will be smooth at the second and second-to-last unique knot."""

    NATURAL = 2
    """The curve will have zero second derivatives at the endpoints."""

    HERMITE = 3
    """Specify the derivatives at the knots."""

    PERIODIC = 4
    """The curve will be periodic at the endpoints."""

    TANGENT = 5
    """Specify the tangents at the endpoints."""

    TANGENTNATURAL = 6
    """Use `TANGENT` for the start and `NATURAL` for the end."""


def line(a, b, relative=False):
    """  Create a line between two points.

    :param array-like a: Start point
    :param array-like b: End point
    :param bool relative: Whether *b* is relative to *a*
    :return: Linear curve from *a* to *b*
    :rtype: Curve
    """
    if relative:
        b = tuple(ai + bi for ai, bi in zip(a, b))
    return Curve(controlpoints=[a, b])


def polygon(*points, **keywords):
    """  Create a linear interpolation between input points.

    :param [array-like] points: The points to interpolate
    :param bool relative: If controlpoints are interpreted as relative to the
        previous one
    :param [float] t: specify parametric interpolation points (the knot vector)
    :return: Linear curve through the input points
    :rtype: Curve
    """
    if len(points) == 1:
        points = points[0]

    knot = keywords.get('t', [])
    if len(knot) == 0: # establish knot vector based on eucledian length between points
        knot = [0, 0]
        prevPt = points[0]
        for pt in points[1:]:
            dist = 0
            for (x0, x1) in zip(prevPt, pt):  # loop over (x,y) and maybe z-coordinate
                dist += (x1 - x0)**2
            knot.append(knot[-1] + sqrt(dist))
            prevPt = pt
        knot.append(knot[-1])
    else: # use knot vector given as input argument
        knot = [knot[0]] + list(knot) + [knot[-1]]

    relative = keywords.get('relative', False)
    if relative:
        points = list(points)
        for i in range(1, len(points)):
            points[i] = [x0 + x1 for (x0,x1) in zip(points[i-1], points[i])]

    return Curve(BSplineBasis(2, knot), points)


def n_gon(n=5, r=1, center=(0,0,0), normal=(0,0,1)):
    """  Create a regular polygon of *n* equal sides centered at the origin.

    :param int n: Number of sides and vertices
    :param float r: Radius
    :param array-like center: local origin
    :param array-like normal: local normal
    :return: A linear, periodic, 2D curve
    :rtype: Curve
    :raises ValueError: If radius is not positive
    :raises ValueError: If *n* < 3
    """
    if r <= 0:
        raise ValueError('radius needs to be positive')
    if n < 3:
        raise ValueError('regular polygons need at least 3 sides')

    cp = []
    dt = 2 * pi / n
    knot = [-1]
    for i in range(n):
        cp.append([r * cos(i * dt), r * sin(i * dt)])
        knot.append(i)
    knot += [n, n+1]
    basis = BSplineBasis(2, knot, 0)

    result =  Curve(basis, cp)
    return flip_and_move_plane_geometry(result, center, normal)

def circle(r=1, center=(0,0,0), normal=(0,0,1), type='p2C0', xaxis=(1,0,0)):
    """  Create a circle.

    :param float r: Radius
    :param array-like center: local origin
    :param array-like normal: local normal
    :param string type: The type of parametrization ('p2C0' or 'p4C1')
    :param array-like xaxis: direction of sem, i.e. parametric start point t=0
    :return: A periodic, quadratic rational curve
    :rtype: Curve
    :raises ValueError: If radius is not positive
    """
    if r <= 0:
        raise ValueError('radius needs to be positive')

    if type == 'p2C0' or type == 'C0p2':
        w = 1.0 / sqrt(2)
        controlpoints = [[1, 0, 1],
                         [w, w, w],
                         [0, 1, 1],
                         [-w, w, w],
                         [-1, 0, 1],
                         [-w, -w, w],
                         [0, -1, 1],
                         [w, -w, w]]
        knot = np.array([-1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5]) / 4.0 * 2 * pi

        result = Curve(BSplineBasis(3, knot, 0), controlpoints, True)
    elif type.lower() == 'p4c1' or type.lower() == 'c1p4':
        w = 2*sqrt(2)/3
        a = 1.0/2/sqrt(2)
        b = 1.0/6 * (4*sqrt(2)-1)
        controlpoints = [[ 1,-a, 1],
                         [ 1, a, 1],
                         [ b, b, w],
                         [ a, 1, 1],
                         [-a, 1, 1],
                         [-b, b, w],
                         [-1, a, 1],
                         [-1,-a, 1],
                         [-b,-b, w],
                         [-a,-1, 1],
                         [ a,-1, 1],
                         [ b,-b, w]]
        knot = np.array([ -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5]) / 4.0 * 2 * pi
        result = Curve(BSplineBasis(5, knot, 1), controlpoints, True)
    else:
        raise ValueError('Unkown type: %s' %(type))

    result *= r
    result.rotate(rotate_local_x_axis(xaxis, normal))
    return flip_and_move_plane_geometry(result, center, normal)

def ellipse(r1=1, r2=1, center=(0,0,0), normal=(0,0,1), type='p2C0', xaxis=(1,0,0)):
    """  Create an ellipse

    :param float r1: Radius along xaxis
    :param float r2: Radius orthogonal to xaxis
    :param array-like center: local origin
    :param array-like normal: local normal
    :param string type: The type of parametrization ('p2C0' or 'p4C1')
    :param array-like xaxis: direction of sem, i.e. parametric start point t=0
    :return: A periodic, quadratic rational curve
    :rtype: Curve
    :raises ValueError: If radius is not positive
    """
    result = circle(type=type)
    result *= [r1,r2,1]
    result.rotate(rotate_local_x_axis(xaxis, normal))
    return flip_and_move_plane_geometry(result, center, normal)

def circle_segment_from_three_points(x0, x1, x2):
    """circle_segment_from_three_points(x0, x1, x2)

    Create a circle segment going from the point x0 to x2 through x1

    :param array-like x0: The start point (2D or 3D point)
    :param array-like x1: An intermediate point (2D or 3D)
    :param array-like x2: The end point (2D or 3D)
    :rtype: Curve
    """
    # wrap input into 3d numpy arrays
    pt0 = np.array([0,0,0], dtype='float')
    pt1 = np.array([0,0,0], dtype='float')
    pt2 = np.array([0,0,0], dtype='float')
    pt0[:len(x0)] = x0
    pt1[:len(x1)] = x1
    pt2[:len(x2)] = x2

    # figure out normal, center and radius
    normal = np.cross(pt1-pt0, pt2-pt0)
    A = np.vstack((2*(pt1-pt0), 2*(pt2-pt0), normal))
    b = np.array([ np.dot(pt1,pt1) - np.dot(pt0,pt0),
                   np.dot(pt2,pt2) - np.dot(pt0,pt0),
                   np.dot(normal,pt0)])
    center = np.linalg.solve(A,b)
    radius = norm(pt2-center)
    v2 = pt2-center
    v1 = pt1-center
    v0 = pt0-center
    w0 = pt0-pt2
    w1 = pt1-pt2
    w2 = np.cross(w0, w1)
    normal = np.cross(v0,v2)
    len_v2 = norm(v2)
    len_v0 = norm(v0)
    theta  = np.arccos(np.dot(v2,v0) / len_v2 / len_v0)
    if not np.all([np.sign(i)==np.sign(j) or abs(i-j) < state.controlpoint_absolute_tolerance for (i,j) in zip(w2,normal)]):
        theta = 2*pi - theta
        normal = -normal

    result = circle_segment(theta, radius, center, np.cross(v0,v1), v0)

    # spit out 2D curve if all input points were 2D, otherwise return 3D
    result.set_dimension(np.max([len(x0), len(x1), len(x2)]))
    return result


def circle_segment(theta, r=1, center=(0,0,0), normal=(0,0,1), xaxis=(1,0,0)):
    """  Create a circle segment starting parallel to the rotated x-axis.

    :param float theta: Angle in radians
    :param float r: Radius
    :param array-like center: circle segment center
    :param array-like normal: normal vector to the plane that contains circle
    :param array-like xaxis: direction of the parametric start point t=0
    :return: A quadratic rational curve
    :rtype: Curve
    :raises ValueError: If radius is not positive
    :raises ValueError: If theta is not in the range *[-2pi, 2pi]*
    """
    # error test input
    if abs(theta) > 2 * pi:
        raise ValueError('theta needs to be in range [-2pi,2pi]')
    if r <= 0:
        raise ValueError('radius needs to be positive')
    if theta == 2*pi:
        return circle(r, center, normal)

    # build knot vector
    knot_spans = int(ceil(abs(theta) / (2 * pi / 3)))
    knot = [0]
    for i in range(knot_spans + 1):
        knot += [i] * 2
    knot += [knot_spans]  # knot vector [0,0,0,1,1,2,2,..,n,n,n]
    knot = np.array(knot) / float(knot[-1]) * theta  # set parametric space to [0,theta]

    n = (knot_spans - 1) * 2 + 3  # number of control points needed
    cp = []
    t = 0  # current angle
    dt = float(theta) / knot_spans / 2  # angle step

    # build control points
    for i in range(n):
        w = 1 - (i % 2) * (1 - cos(dt))  # weights = 1 and cos(dt) every other i
        x = r * cos(t)
        y = r * sin(t)
        cp += [[x, y, w]]
        t += dt

    if theta < 0:
        cp.reverse()
        result = Curve(BSplineBasis(3, np.flip(knot,0)), cp, True)
    else:
        result = Curve(BSplineBasis(3, knot), cp, True)
    result.rotate(rotate_local_x_axis(xaxis, normal))
    return flip_and_move_plane_geometry(result, center, normal)

def interpolate(x, basis, t=None):
    """  Perform general spline interpolation on a provided basis.

    :param matrix-like x: Matrix *X[i,j]* of interpolation points *xi* with
        components *j*
    :param BSplineBasis basis: Basis on which to interpolate
    :param array-like t: parametric values at interpolation points; defaults to
        Greville points if not provided
    :return: Interpolated curve
    :rtype: Curve
    """
    # wrap input into an array
    x = np.array(x)

    # evaluate all basis functions in the interpolation points
    if t is None:
        t = basis.greville()
    N = basis.evaluate(t, sparse=True)

    # solve interpolation problem
    cp = splinalg.spsolve(N, x)
    cp = cp.reshape(x.shape)

    return Curve(basis, cp)

def least_square_fit(x, basis, t):
    """  Perform a least-square fit of a point cloud onto a spline basis

    :param matrix-like x: Matrix *X[i,j]* of interpolation points *xi* with
        components *j*. The number of points must be equal to or larger than
        the number of basis functions in *basis*
    :param BSplineBasis basis: Basis on which to interpolate
    :param array-like t: parametric values at evaluation points
    :return: Approximated curve
    :rtype: Curve
    """

    # evaluate all basis functions at evaluation points
    N = basis.evaluate(t)

    # solve interpolation problem
    controlpoints,_,_,_ = np.linalg.lstsq(N, x, rcond=None)

    return Curve(basis, controlpoints)


def cubic_curve(x, boundary=Boundary.FREE, t=None, tangents=None):
    """  Perform cubic spline interpolation on a provided basis.

    The valid boundary conditions are enumerated in :class:`Boundary`. The
    meaning of the `tangents` parameter depends on the specified boundary
    condition:

    - `TANGENT`: two points,
    - `TANGENTNATURAL`: one point,
    - `HERMITE`: *n* points

    :param matrix-like x: Matrix *X[i,j]* of interpolation points *x_i* with
        components *j*
    :param int boundary: Any value from :class:`Boundary`.
    :param array-like t: parametric values at interpolation points, defaults
        to Euclidean distance between evaluation points
    :param matrix-like tangents: Tangent information according to the boundary
        conditions.
    :return: Interpolated curve
    :rtype: Curve
    """

    # if periodic input is not closed, make sure we do it now
    if boundary == Boundary.PERIODIC and not (
       np.allclose(x[0,:], x[-1,:],
                   rtol = state.controlpoint_relative_tolerance,
                   atol = state.controlpoint_absolute_tolerance)):
        x = np.append(x, [x[0,:]], axis=0)
        if t is not None:
            # augment interpolation knot by euclidian distance to end
            t = list(t) + [t[-1] + norm(np.array(x[0,:])- np.array(x[-2,:]))]

    n = len(x)
    if t is None:
        t = [0.0]
        for (x0,x1) in zip(x[:-1,:], x[1:,:]):
            # eucledian distance between two consecutive points
            dist = norm(np.array(x1)-np.array(x0))
            t.append(t[-1]+dist)

    # modify knot vector for chosen boundary conditions
    knot = [t[0]]*3 + list(t) + [t[-1]]*3
    if boundary == Boundary.FREE:
        del knot[-5]
        del knot[4]
    elif boundary == Boundary.HERMITE:
        knot = sorted(list(knot) + list(t[1:-1]))

    # create the interpolation basis and interpolation matrix on this
    if boundary == Boundary.PERIODIC:
        # C2-periodic knots
        knot[0]  = t[0]  + t[-4] - t[-1]
        knot[1]  = t[0]  + t[-3] - t[-1]
        knot[2]  = t[0]  + t[-2] - t[-1]
        knot[-3] = t[-1] + t[1]  - t[0]
        knot[-2] = t[-1] + t[2]  - t[0]
        knot[-1] = t[-1] + t[3]  - t[0]

        basis = BSplineBasis(4, knot, 2)

        # do not duplicate the interpolation at the sem (start=end is the same point)
        # identical points equal singular interpolation matrix
        t = t[:-1]
        x = x[:-1,:]
    else:
        basis = BSplineBasis(4, knot)
    N = basis(t, sparse=True)  # left-hand-side matrix

    # add derivative boundary conditions if applicable
    if boundary in [Boundary.TANGENT, Boundary.HERMITE, Boundary.TANGENTNATURAL]:
        if boundary == Boundary.TANGENT:
            dn = basis([t[0], t[-1]], d=1)
        elif boundary == Boundary.TANGENTNATURAL:
            dn = basis(t[0], d=1)
        elif boundary == Boundary.HERMITE:
            dn = basis(t, d=1)
        N  = sp.vstack([N, sp.csr_matrix(dn)])
        x  = np.vstack([x, tangents])

    # add double derivative boundary conditions if applicable
    if boundary in [Boundary.NATURAL, Boundary.TANGENTNATURAL]:
        if boundary == Boundary.NATURAL:
            ddn  = basis([t[0], t[-1]], d=2)
            new  = 2
        elif boundary == Boundary.TANGENTNATURAL:
            ddn  = basis(t[-1], d=2)
            new  = 1
        N  = sp.vstack([N, sp.csr_matrix(ddn)])
        x  = np.vstack([x, np.zeros((new, x.shape[1]))])

    # solve system to get controlpoints
    cp = splinalg.spsolve(N,x)
    cp = cp.reshape(x.shape)

    # wrap it all into a curve and return
    return Curve(basis, cp)

def bezier(pts, quadratic=False, relative=False):
    """  Generate a cubic or quadratic bezier curve from a set of control points

    :param [array-like] pts: list of control-points. In addition to a starting
        point we need three points per bezier interval for cubic splines and
        two points for quadratic splines
    :param bool quadratic: True if a quadratic spline is to be returned, False
        if a cubic spline is to be returned
    :param bool relative: If controlpoints are interpreted as relative to the
        previous one
    :return: Bezier curve
    :rtype: Curve

    """
    if quadratic:
        p = 3
    else:
        p = 4
    # compute number of intervals
    n = int((len(pts)-1)/(p-1))
    # generate uniform knot vector of repeated integers
    knot = list(range(n+1)) * (p-1) + [0, n]
    knot.sort()
    if relative:
        pts = copy.deepcopy(pts)
        for i in range(1, len(pts)):
            pts[i] = [x0 + x1 for (x0,x1) in zip(pts[i-1], pts[i])]
    return Curve(BSplineBasis(p, knot), pts)

def manipulate(crv, f, normalized=False, vectorized=False):
    """  Create a new curve based on an expression-evaluation of an existing one
    :param Curve crv: original curve on which f is to be applied
    :param function f: expression of the physical point *x*, the velocity
        (tangent) *v*, parametric point *t* and/or acceleration *a*.
    :param normalized: If velocity and acceleration terms should be normalized
        to have length 1
    :param vectorized: True if *f* is expressed in terms of vectorized
        operations. The method is sped up whenever this is used.

    Examples:

    .. code:: python

        def scale_by_two(x):
            return 2*x

        new_curve = manipulate(old_curve, scale_by_two)
        new_curve = old_curve * 2 # will give the same result

        def move_3_to_right(x, v):
            result = x
            # *v* is velocity. Rotate this 90 to the right and it points (+v[1], -v[0])
            result[0] += 3*v[1]
            result[1] -= 3*v[0]
            return result

        # note that the velocity vector will not have length one unless normalized is passed
        new_curve = manipulate(old_curve, move_3_to_right, normalized=True)

        def move_3_to_right_fast(x, v):
            result = x
            result[:,0] += 3*v[:,1]
            result[:,1] -= 3*v[:,0]
            return result

        new_curve = manipulate(old_curve, move_3_to_right_fast, normalized=True, vectorized=True)
    """

    b = crv.bases[0]
    t = np.array(b.greville())
    n = len(crv)

    if vectorized:
        x = crv(t)
        arg_names = inspect.getargspec(f).args
        argc = len(arg_names)
        argv = [0] * argc
        for j in range(argc):
            if arg_names[j] == 'x':
                argv[j] = x
            elif arg_names[j] == 't':
                argv[j] = t
            elif arg_names[j] == 'v':
                c0 = np.array([i for i in range(n) if b.continuity(t[i]) == 0])
                v = crv.derivative(t, 1)
                if len(c0)>0:
                    v[c0,:] = (v[c0,:] + crv.derivative(t[c0], 1, above=False)) / 2.0
                if normalized:
                    v[:] = [vel / norm(vel) for vel in v]
                argv[j] = v
            elif arg_names[j] == 'a':
                c1 = np.array([i for i in range(n) if b.continuity(t[i]) < 2])
                a = crv.derivative(t, 2)
                if len(c1)>0:
                    a[c1,:] = (a[c1,:] + crv.derivative(t[c1], 2, above=False)) / 2.0
                if normalized:
                    a[:] = [acc / norm(acc) for acc in a]
                argv[j] = a
        destination   = f(*argv)
    else:
        destination = np.zeros((len(crv), crv.dimension))
        for (t1, i) in zip(t, range(len(t))):
            x = crv(t1)
            arg_names = inspect.getargspec(f).args
            argc = len(arg_names)
            argv = [0] * argc
            for j in range(argc):
                if arg_names[j] == 'x':
                    argv[j] = x
                elif arg_names[j] == 't':
                    argv[j] = t1
                elif arg_names[j] == 'v':
                    v = crv.derivative(t1, 1)
                    if b.continuity(t1) < 1:
                        v += crv.derivative(t1, 1, above=False)
                        v /= 2.0
                    if normalized:
                        v /= norm(v)
                    argv[j] = v
                elif arg_names[j] == 'a':
                    a = crv.derivative(t1, 2)
                    if b.continuity(t1) < 2:
                        a += crv.derivative(t1, 1, above=False)
                        a /= 2.0
                    if normalized:
                        a /= norm(a)
                    argv[j] = a
            destination[i] = f(*argv)

    N = b(t, sparse=True)
    controlpoints = splinalg.spsolve(N, destination)
    return Curve(b, controlpoints)

def fit(x, t0, t1, rtol=1e-4, atol=0.0):
    """  Computes an interpolation for a parametric curve up to a specified tolerance.
    The method will iteratively refine parts where needed resulting in a non-uniform
    knot vector with as optimized knot locations as possible.

    :param function x: callable function which takes as input a vector of evaluation
        points t and gives as output a matrix x where x[i,j] is component j evaluated
        at point t[i]
    :param float t0: start of parametric domain
    :param float t1: end of parametric domain
    :param float rtol: relative tolerance for stopping criterium. It is defined
        to be ||e||_L2 / D, where D is the length of the curve and ||e||_L2 is the
        L2-error (see Curve.error)
    :param float atol: absolute tolerance for stopping criterium. It is defined to
        be the maximal distance between the curve approximation and the exact curve
    :return: Curve Non-uniform cubic B-spline curve

    Examples:

    .. code:: python

        import numpy as np
        import splipy.curve_factory as curve_factory

        # gives a B-spline approximation to the circle with arclength parametrization
        # unlike curve_factory.circle which is exact, but not arclength
        def arclength_circle(t):
            return np.array( [np.cos(t), np.sin(t)] ).T
        crv = curve_factory.fit(arclength_circle, 0, 2*np.pi)
        print(crv)

        # approximates a difficult function with wild behaviour around t=0, but
        # this is overcome by a higher knot density around this point
        def one_over_t(t):
            eps = 1e-8 # to avoid 1/0 we add a small epsilon
            return np.array( [t, 1.0/(t+eps)] ).T
        crv = curve_factory.fit(one_over_t, 0, 1, rtol=1e-6)
        print(crv) # first knot span is ~1e-9, last knot is ~1e-1

        # one can specify the target curve in terms of existing Curve objects
        crv = curve_factory.circle(r=1)     # Curve-object, quadratic NURBS
        def move_along_tangent(t):
            return crv(t) + crv.tangent(t)  # can evaluate curve or its derivatives
        # fit() will create a B-spline approximation using non-uniform refinement
        crv2 = curve_factory.fit(move_along_tangent, crv.start(0), crv.end(0))
    """

    b = BSplineBasis(4, [t0,t0,t0,t0, t1,t1,t1,t1])
    t = np.array(b.greville())
    crv = interpolate(x(t), b, t)
    (err2, maxerr) = crv.error(x)
    # polynomial input (which can be exactly represented) only use one knot span
    if maxerr < 1e-13:
        return crv

    # for all other curves, start with 4 knot spans
    knot_vector = [t0,t0,t0,t0] + [i/5.0*(t1-t0)+t0 for i in range(1,5)] + [t1,t1,t1,t1]
    b = BSplineBasis(4, knot_vector)
    t = np.array(b.greville())
    crv = interpolate(x(t), b, t)
    (err2, maxerr) = crv.error(x)
    # this is technically false since we need the length of the target function *x*
    # and not our approximation *crv*, but we don't have the derivative of *x*, so
    # we can't compute it. This seems like a healthy compromise
    length = crv.length()
    while np.sqrt(np.sum(err2))/length > rtol and maxerr > atol:
        knot_span    = crv.knots(0) # knot vector without multiplicities
        target_error = (rtol*length)**2 / len(err2) # equidistribute error among all knot spans
        refinements  = []
        for i in range(len(err2)):
            # figure out how many new knots we require in this knot interval:
            # if we converge with *scale* and want an error of *target_error*
            # |e|^2 * (1/n)^scale = target_error^2

            conv_order = 4                   # cubic interpolateion is order=4
            square_conv_order = 2*conv_order # we are computing with square of error
            scale = square_conv_order + 4    # don't want to converge too quickly in case of highly non-uniform mesh refinement is required
            n = int(np.ceil(np.exp((np.log(err2[i]) - np.log(target_error))/scale)))

            # add *n* new interior knots to this knot span
            new_knots = np.linspace(knot_span[i], knot_span[i+1], n+1)
            knot_vector = knot_vector + list(new_knots[1:-1])

        # build new refined knot vector
        knot_vector.sort()
        b = BSplineBasis(4, knot_vector)
        # do interpolation and return result
        t = np.array(b.greville())
        crv = interpolate(x(t), b, t)
        (err2, maxerr) = crv.error(x)
        length = crv.length()

    return crv

def fit_points(x, t=[], rtol=1e-4, atol=0.0):
    """ Computes an approximation for a list of points up to a specified tolerance.
    The method will iteratively refine parts where needed resulting in a non-uniform
    knot vector with as optimized knot locations as possible. The target curve is the
    linear interpolation of the input points

    :param [point-like] x: The points to approximate
    :param [float] t: parametric values for curve points (same length as x)
    :param float rtol: relative tolerance for stopping criterium. It is defined
        to be ||e||_L2 / D, where D is the length of the curve and ||e||_L2 is the
        L2-error (see Curve.error)
    :param float atol: absolute tolerance for stopping criterium. It is defined to
        be the maximal distance between the curve approximation and the exact curve
    :return: Curve Non-uniform cubic B-spline curve
    """

    if len(t)>0:
        linear = polygon(x, t=t)
    else:
        linear = polygon(x)
    return fit(linear, linear.start(0), linear.end(0), rtol=rtol, atol=atol)



"""This module handles the global Splipy state."""

from contextlib import contextmanager
import sys

states = ['controlpoint_relative_tolerance',
          'controlpoint_absolute_tolerance',
          'parametric_relative_tolerance',
          'parametric_absolute_tolerance',
          'knot_tolerance',
          'unlimited']
__all__ = states + ['state']


controlpoint_absolute_tolerance = 1e-8
"""Absolute tolerance used for matching control points."""

controlpoint_relative_tolerance = 0.0
"""Relative tolerance used for matching control points."""

parametric_absolute_tolerance = 1e-8
"""Absolute tolerance used for matching parametric points."""

parametric_relative_tolerance = 0.0
"""Relative tolerance used for matching parametric points."""

knot_tolerance = 1e-10
"""Absolute tolerance used for matching parametric values (knot vectors)."""

unlimited = 1e4
"""Since splipy insists on finite parametric domains, we define 'unbounded' here"""


@contextmanager
def state(**kwargs):
    """A context manager for running code in a modified state.

    This takes an arbitrary number of keyword arguments, which correspond to
    state variables. The managed block runs with these values, which are then
    restored to their previous values afterwards.

    For example:

    .. code:: python

       from splipy.state import state

       with state(controlpoint_relative_tolerance=1e-2):
           # Code that runs with relative tolerance 1e-2
       # Relative tolerance restored at this point
    """
    module = sys.modules[__name__]

    # Store the current values
    before = {k: getattr(module, k) for k in states}

    # Set values according to the input
    for k, v in kwargs.items():
        setattr(module, k, v)

    yield

    # Return settings to their previous values
    for k, v in before.items():
        setattr(module, k, v)



# -*- coding: utf-8 -*-

from collections import Counter, OrderedDict, namedtuple
from itertools import chain, product, permutations, islice

import numpy as np

from .splineobject import SplineObject
from .utils import check_section, sections, section_from_index, section_to_index, uniquify, is_right_hand
from . import state

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping


def _section_to_index(section):
    """Replace all `None` in `section` with `slice(None)`, so that it
    works as a numpy array indexing tuple.
    """
    return tuple(slice(None) if s is None else s for s in section)


face_t = np.dtype([('nodes', int, (4,)), ('owner', int, ()), ('neighbor', int, ()), ('name', object, ())])


class VertexDict(MutableMapping):
    """A dictionary where the keys are numpy arrays, and where equality
    is computed in an approximate sense for floating point numbers.

    All keys must have the same dimensions.
    """

    def __init__(self):
        # List of (key, value) pairs
        self.internal = []

    def _eq(self, a, b):
        """Check whether two numpy arrays are almost equal, according to the given
        tolerances.
        """
        return np.allclose(a, b,
                           rtol=state.controlpoint_relative_tolerance,
                           atol=state.controlpoint_absolute_tolerance)

    def _candidate(self, key):
        """Return the internal index for the first stored mapping that matches the
        given key.

        :param numpy.array key: The key to look for
        :raises KeyError: If the key is not found
        """
        for i, (k, v) in enumerate(self.internal):
            if self._eq(key, k):
                return i
        raise KeyError(key)

    def __setitem__(self, key, value):
        """Assign a key to a value."""
        try:
            c = self._candidate(key)
            k, _ = self.internal[c]
            self.internal[c] = (k, value)
        except KeyError:
            self.internal.append((key, value))

    def __getitem__(self, key):
        """Gets the value assigned to a key.

        :raises KeyError: If the key is not found
        """
        c = self._candidate(key)
        return self.internal[c][1]

    def __delitem__(self, key):
        """Deletes an assignment."""
        self.internal = [(k, v) for k, v in self.internal
                         if not self._eq(key, k)]

    def __iter__(self):
        """Iterate over all keys.

        .. note:: This generates all the stored keys, not all matching keys.
        """
        for k, _ in self.internal:
            yield k

    def items(self):
        """Return a list of key, value pairs."""
        return list(self.internal)

    def __len__(self):
        """Returns the number of stored assignments."""
        return len(self.internal)


class OrientationError(RuntimeError):
    """An `OrientationError` is raised by certain methods in
    :class:`splipy.SplineModel.Orientation` indicating an inability to match
    two objects.
    """
    pass

class TwinError(RuntimeError):
    """A `TwinError` is raised when two objects with identical interfaces
    are added, but different interiors.
    """
    pass

class Orientation(object):
    """An `Orientation` represents a mapping between two coordinate systems: the
    *reference* system and the *actual* or *mapped* system.

    The orientation is represented in terms of two tuples:

    - `perm` is a permutation: direction `d` in the reference system is mapped
      to direction `perm[d]` in the actual system.
    - `flip` is a tuple of bools: `flip[d]` will be `True` to indicate that
      direction `d` *in the reference system* should be reversed.
    """

    def __init__(self, perm, flip):
        """Initialize an Orientation object.

        .. warning:: This constructor is for internal use. Use
            :func:`splipy.SplineModel.Orientation.compute` to generate new
            orientation objects.
        """
        self.perm = perm
        self.flip = flip
        self.perm_inv = tuple(perm.index(d) for d in range(len(perm)))

    @classmethod
    def compute(cls, cpa, cpb=None):
        """Compute and return a new orientation object representing the mapping between
        `cpa` (the reference system) and `cpb` (the mapped system).

        If `cpb` is not given, the identity orientation is returned.

        :param SplineObject cpa: The reference system
        :param SplineObject cpb: The mapped system
        :return: The orientation relating the two systems
        :rtype: Orientation
        :raises OrientationError: If the two objects do not match
        """

        pardim = cpa.pardim

        # Return the identity orientation if no cpb
        if cpb is None:
            return cls(tuple(range(pardim)),
                       tuple(False for _ in range(pardim)))

        # Deal with the easy cases: dimension mismatch, and
        # comparing the shapes as multisets
        if cpa.pardim != cpb.pardim:
            raise OrientationError("Mismatching parametric dimensions")
        if cpa.dimension != cpb.dimension:
            raise OrientationError("Mismatching physical dimensions")
        if Counter(cpa.shape) != Counter(cpb.shape):
            raise OrientationError("Non-matching objects (different shape)")

        cps_a = cpa.controlpoints
        cps_b = cpb.controlpoints

        # If one object is rational and the other is not, promote the
        # non-rational CP array to a rational one
        if cpa.rational and not cpb.rational:
            cps_b = np.concatenate((cps_b, np.ones((*cps_b.shape[:-1], 1))), axis=-1)
        elif cpb.rational and not cpa.rational:
            cps_a = np.concatenate((cps_a, np.ones((*cps_a.shape[:-1], 1))), axis=-1)

        # At this point, both CP arrays are rational or non-rational.
        # If any are rational, remove the weight scaling degree of
        # freedom. We must copy to avoid clobbering the original data.
        if cpa.rational or cpb.rational:
            cps_a = cps_a.copy()
            cps_a[..., -1] /= np.sum(cps_a[..., -1])
            cps_b = cps_b.copy()
            cps_b[..., -1] /= np.sum(cps_b[..., -1])

        # Enumerate all permutations of directions
        for perm in permutations(range(pardim)):
            transposed = cps_b.transpose(perm + (pardim,))
            if transposed.shape != cps_a.shape:
                continue
            # Enumerate all possible direction reversals
            for flip in product([False, True], repeat=pardim):
                slices = tuple(slice(None, None, -1) if f else slice(None) for f in flip)
                test_b = transposed[slices + (slice(None),)]
                if np.allclose(cps_a, test_b,
                               rtol=state.controlpoint_relative_tolerance,
                               atol=state.controlpoint_absolute_tolerance):
                    if all([cpa.bases[i].matches(cpb.bases[perm[i]], reverse=flip[i]) for i in range(pardim)]):
                        return cls(perm, flip)

        raise OrientationError("Non-matching objects")

    @property
    def pardim(self):
        return len(self.perm)

    def __mul__(self, other):
        """Compose two mappings.

        If `ort_left` maps system `A` (reference) to system `B`, and
        `ort_right` maps system `B` (reference) to system `C`, then `ort_left *
        ort_right` maps system `A` (reference) to system `C`.
        """
        assert self.pardim == other.pardim
        perm = tuple(other.perm[self.perm[d]] for d in range(self.pardim))

        # Flip a direction if it is flipped in exactly one orientation
        flip = tuple(self.flip[d] != other.flip[self.perm[d]] for d in range(self.pardim))

        return Orientation(perm, flip)

    def map_array(self, array):
        """Map an array in the mapped system to the reference system."""
        array = array.transpose(*self.perm)
        flips = tuple(slice(None, None, -1) if f else slice(None) for f in self.flip)
        return array[flips]

    def map_section(self, section):
        """Map a section in the mapped system to the reference system.

        The input is a section tuple as described in
        :func:`splipy.Utils.section`. It should be described relative to the
        mapped coordinate system. The return value is the corresponding section
        relative to the reference system.
        """
        permuted = tuple(section[d] for d in self.perm)

        flipped = ()
        for s, f, in zip(permuted, self.flip):
            # Flipping only applies to indexed directions, not variable ones
            if f and s is not None:
                flipped += (0 if s == -1 else -1,)
            else:
                flipped += (s,)

        return flipped

    def view_section(self, section):
        """Reduce a mapping to a lower dimension.

        The input is a section tuple as described in
        :func:`splipy.Utils.section`. It should be described relative to the
        mapped coordinate system. The return value is an `Orientation` object
        that describes the same orentation, relative to the section only.

        For example, for a 2D orentation that flips only the second direction,
        `view_section` will return the identity orientation for edges along the
        first direction, but the inverse orientation for edges along the second
        (flipped) direction.
        """
        # The directions (in the mapped system) that are variable
        variable_dirs = [i for i, s in enumerate(section) if s is None]

        # The directions (in the reference system) that are variable
        # Sort to get the order that they have in the reference system
        actual_dirs = sorted(self.perm_inv[d] for d in variable_dirs)

        # perm[d] maps a direction to the mapped system, and its index in
        # variable_dirs reveals the permutation
        new_perm = tuple(variable_dirs.index(self.perm[d]) for d in actual_dirs)

        new_flip = tuple(self.flip[d] for d in actual_dirs)
        return self.__class__(new_perm, new_flip)

    @property
    def ifem_format(self):
        """Compute the orientation in IFEM format.

        For one-dimensional objects, this is a single binary digit indicating
        if the direction is reversed or not.

        For two-dimensional objects, this is a single-digit octal number with
        binary digits `ijk` with the following meanings:

        - `i` is 1 if the directions are swapped
        - `j` is 1 if the first direction in the reference system should be
          reversed.
        - `k` is 1 if the second direction in the reference system should be
          reversed.

        :raises RuntimeError: If the parametric dimension is not supported.
        """
        if len(self.flip) == 0:
            return 0
        if len(self.flip) == 1:
            return int(self.flip[0])
        elif len(self.flip) == 2:
            ret = 0
            for i, axis in enumerate(self.perm[::-1]):
                if self.flip[axis]:
                    ret |= 1 << i
            if tuple(self.perm) == (1,0):
                ret |= 1 << 2
            return ret
        raise RuntimeError(
            'IFEM orientation format not supported for pardim {}'.format(len(self.flip))
        )


class TopologicalNode(object):
    """A `TopologicalNode` object refers to a single, persistent point in the
    topological graph. It represents some object of dimension `d` (that is, a
    point, an edge, etc.) and it has references to all the other objects it
    touches (that is, a face has a reference to four vertices, four edges, and
    any volumes it may be part of).

    Relevant attributes are:

    - obj: The underlying `SplineObject`
    - lower_nodes: A nested list of lower-order connections to other
      `TopologicalNode` objects. `lower_nodes[d]` is a list of all connected
      node objects with dimension `d`.
    - higher_nodes: A dictionary of higher-order connections to other
      `TopologicalNode` objects. `higher_nodes[d]` is a list of all connected
      node objects with dimension `d`.
    - owner: A top-level `TopologicalNode` object that "owns" this node.

    .. note:: Connections to lower order nodes are `ordered` corresponding to
        the natural ordering of sections (see :func:`splipy.Utils.sections`).
        Connections to higher order nodes are not ordered.

    .. warning:: This class is mostly for internal use. It performs no checks
        of any kind.
    """

    def __init__(self, obj, lower_nodes, index):
        """Initialize a `TopologicalNode` object associated with the given
        `SplineObject` and lower order nodes.

        :param SplineObject obj: The underlying spline object
        :param lower_nodes: A nested list of lower order nodes
        :param index: The node number
        """
        self.obj = obj
        self.lower_nodes = lower_nodes
        self.index = index
        self.higher_nodes = {}
        self.owner = None

        self.name = None
        self.cell_numbers = None
        self.cp_numbers = None

        for dim_nodes in self.lower_nodes:
            for node in dim_nodes:
                node.assign_higher(self)

        # Take ownership of lower nodes that are unaccounted for
        if self.pardim > 0:
            for node in lower_nodes[-1]:
                if node.owner is None:
                    node._transfer_ownership(self)

    @property
    def pardim(self):
        return self.obj.pardim

    @property
    def nhigher(self):
        return len(self.higher_nodes[self.pardim + 1])

    @property
    def super_owner(self):
        """Return the highest owning node."""
        owner = self
        while owner.owner is not None:
            owner = owner.owner
        return owner

    def assign_higher(self, node):
        """Add a link to a node of higher dimension."""
        self.higher_nodes.setdefault(node.pardim, list()).append(node)

    def view(self, other_obj=None):
        """Return a `NodeView` object of this node.

        The returned view has an orientation that matches that of the input
        argument. If the argument is omitted, the identity orientation is used.

        :param SplineObject other_obj: The "view" object
        :raises OrientationError: If the object does not match this node's
            underlying object
        """
        if other_obj:
            orientation = Orientation.compute(self.obj, other_obj)
        else:
            orientation = Orientation.compute(self.obj)
        return NodeView(self, orientation)

    def _transfer_ownership(self, new_owner):
        """Transfers ownership of this node to a new owner. This operation is
        transitive, so all child nodes owned by this node, or who are
        owner-less will also be transferred.

        :param TopologicalNode new_owner: The new owner
        """
        self.owner = new_owner

        if self.pardim > 0:
            for child in self.lower_nodes[-1]:
                if child.owner is self or child.owner is None:
                    child._transfer_ownership(new_owner)

    def generate_cp_numbers(self, start=0):
        """Generate a control point numbering starting at `start`. Return the next unused index."""
        assert self.owner is None

        # Initialize a number array
        shape = self.obj.shape
        numbers = np.empty(shape, dtype=int)
        numbers[:] = 0

        # Flag control points owned by other top-level objects with -1
        for node, section in zip(self.lower_nodes[-1], sections(self.pardim, self.pardim-1)):
            if node.owner is not self:
                numbers[_section_to_index(section)] = -1

        # Fill in control point numbers for the ones we do own
        mask = np.where(numbers != -1)
        nowned = len(mask[0])
        numbers[mask] = np.arange(start, start + nowned, dtype=int)

        # This method takes care of communicating results to children
        self.assign_cp_numbers(numbers)
        return start + nowned

    def assign_cp_numbers(self, numbers):
        """Directly assign control point numbers."""
        self.cp_numbers = numbers

        # Control point numbers for owned children must be communicated to them
        if self.pardim > 0:
            for node, section in zip(self.lower_nodes[-1], sections(self.pardim, self.pardim-1)):
                if node.owner is self or node.owner is self.owner:
                    # Since this runs in a direct line of ownership, we don't need to be concerned with
                    # orientations not matching up.
                    node.assign_cp_numbers(numbers[_section_to_index(section)])

    def read_cp_numbers(self):
        """Read control point numbers for unowned control points from child nodes."""
        for node, section in zip(self.lower_nodes[-1], sections(self.pardim, self.pardim-1)):
            if node.owner is not self:
                # The two sections may not agree on orientation, so we fix this here.
                ori = Orientation.compute(self.obj.section(*section), node.obj)
                self.cp_numbers[_section_to_index(section)] = ori.map_array(node.cp_numbers)

        assert (self.cp_numbers != -1).all()

    def generate_cell_numbers(self, start=0):
        """Generate a cell numbering starting at `start`. Return the next unused index."""
        assert self.owner is None

        # Cells are never shared between top-level objects, so no need to worry about ownership here
        shape = [len(kvec) - 1 for kvec in self.obj.knots()]
        nelems = np.prod(shape)
        self.cell_numbers = np.reshape(np.arange(start, start + nelems, dtype=int), shape)
        return start + nelems

    def faces(self):
        """Return all faces owned by this node, as a list of numpy arrays with dtype `face_t`."""
        assert self.pardim == 3
        assert self.obj.order() == (2,2,2)
        shape = [len(kvec) - 1 for kvec in self.obj.knots()]
        ncells = np.prod(shape)
        retval = []

        def mkindex(dim, z, a, b):
            rval = [a, b] if dim != 1 else [b, a]
            rval.insert(dim, z)
            return tuple(rval)

        lower = iter(self.lower_nodes[-1])
        for d in range(self.pardim):
            # Number of faces in one "slice"
            nperslice = ncells // shape[d]

            # First, get all internal faces in this direction
            # The owner (lowest cell index) is guaranteed to be toward the lower end
            # TODO: We assume a right-hand coordinate system here
            nfaces = ncells - nperslice
            faces = np.empty((nfaces,), dtype=face_t)
            faces['nodes'][:,0] = self.cp_numbers[mkindex(d, np.s_[1:-1], np.s_[:-1], np.s_[:-1])].flatten()
            faces['nodes'][:,1] = self.cp_numbers[mkindex(d, np.s_[1:-1], np.s_[1:],  np.s_[:-1])].flatten()
            faces['nodes'][:,2] = self.cp_numbers[mkindex(d, np.s_[1:-1], np.s_[1:],  np.s_[1:])].flatten()
            faces['nodes'][:,3] = self.cp_numbers[mkindex(d, np.s_[1:-1], np.s_[:-1], np.s_[1:])].flatten()
            faces['owner'] = self.cell_numbers[mkindex(d, np.s_[:-1], np.s_[:], np.s_[:])].flatten()
            faces['neighbor'] = self.cell_numbers[mkindex(d, np.s_[1:],  np.s_[:], np.s_[:])].flatten()
            retval.append(faces)

            # Go through the two boundaries
            for bdnode, bdindex in zip(islice(lower, 2), (0, -1)):
                assert bdnode.nhigher in {1, 2}

                # Faces on an interface are only returned from the owner
                if bdnode.owner is not self:
                    continue

                faces = np.empty((nperslice,), dtype=face_t)
                faces['nodes'][:,0] = self.cp_numbers[mkindex(d, bdindex, np.s_[:-1], np.s_[:-1])].flatten()
                faces['nodes'][:,1] = self.cp_numbers[mkindex(d, bdindex, np.s_[1:], np.s_[:-1])].flatten()
                faces['nodes'][:,2] = self.cp_numbers[mkindex(d, bdindex, np.s_[1:], np.s_[1:])].flatten()
                faces['nodes'][:,3] = self.cp_numbers[mkindex(d, bdindex, np.s_[:-1], np.s_[1:])].flatten()
                faces['owner'] = self.cell_numbers[mkindex(d, bdindex, np.s_[:], np.s_[:])].flatten()
                faces['name'] = bdnode.name

                # If we're on the left boundary, the face normal must point in the other direction
                # NOTE: We copy when swapping here, since we are swapping values which are views into
                # a mutable array!
                if bdindex == 0:
                    faces['nodes'][:,1], faces['nodes'][:,3] = (
                        faces['nodes'][:,3].copy(), faces['nodes'][:,1].copy()
                    )

                # If there's a neighbor on the interface we need neighbouring cell numbers
                if bdnode.nhigher == 1:
                    faces['neighbor'] = -1
                else:
                    neighbor = next(c for c in bdnode.higher_nodes[3] if c is not self)

                    # Find out which face the interface is as numbered from the neighbor's perspective
                    nb_index = neighbor.lower_nodes[2].index(bdnode)

                    # Get the spline object on that interface as oriented from the neighbor's perspective
                    nb_sec = section_from_index(3, 2, nb_index)
                    nb_obj = neighbor.obj.section(*nb_sec)

                    # Compute the relative orientation
                    ori = Orientation.compute(bdnode.obj, nb_obj)

                    # Get the neighbor cell numbers from the neighbor's perspective, and map them to our system
                    cellidxs = neighbor.cell_numbers[_section_to_index(nb_sec)]
                    faces['neighbor'] = ori.map_array(cellidxs).flatten()

                retval.append(faces)

        for faces in retval:
            assert ((faces['owner'] < faces['neighbor']) | (faces['neighbor'] == -1)).all()
        return retval


class NodeView(object):
    """A `NodeView` object refers to a *view* to a point in the topological graph.
    It is composed of a node (:class:`splipy.SplineModel.TopologicalNode`) and
    an orientation (:class:`splipy.SplineModel.Orienation`).

    .. note:: Unlike `TopologicalNode` objects, `NodeView` objects are not
        persistent.
    """

    def __init__(self, node, orientation=None):
        """Initialize a `NodeView` object with the given node and orientation.

        .. warning:: This constructor is for internal use.
        """
        self.node = node
        self.orientation = orientation

    @property
    def pardim(self):
        return self.node.pardim

    @property
    def name(self):
        return self.node.name

    @name.setter
    def name(self, value):
        self.node.name = value

    def section(self, *args, **kwargs):
        """Return a section. See :func:`splipy.SplineObject.section` for more details
        on the input arguments.

        The return value is another `NodeView` object with a different
        underlying node, but whose orientation corresponds to the original
        view, as expected.
        """
        section = check_section(*args, pardim=self.pardim, **kwargs)
        tgt_dim = sum(1 for s in section if s is None)

        # The index of the section in the reference system
        ref_idx = section_to_index(self.orientation.map_section(section))

        # The underlying node
        node = self.node.lower_nodes[tgt_dim][ref_idx]

        # The underlying lower-order node may not have an orientation that
        # matches the higher-order node, so we need to compose two orientations
        ref_ori = Orientation.compute(node.obj, self.node.obj.section(*section, unwrap_points=False))
        my_ori = self.orientation.view_section(section)

        return NodeView(node, ref_ori * my_ori)

    def corner(self, i):
        """Return the i'th corner."""
        return self.section(*section_from_index(self.pardim, 0, i))

    @property
    def corners(self):
        """A tuple of all corners."""
        return tuple(self.section(s) for s in sections(self.pardim, 0))

    def edge(self, i):
        """Return the i'th edge."""
        return self.section(*section_from_index(self.pardim, 1, i))

    @property
    def edges(self):
        """A tuple of all edges."""
        return tuple(self.section(s) for s in sections(self.pardim, 1))

    def face(self, i):
        """Return the i'th face."""
        return self.section(*section_from_index(self.pardim, 2, i))

    @property
    def faces(self):
        """A tuple of all faces."""
        return tuple(self.section(s) for s in sections(self.pardim, 2))


class ObjectCatalogue(object):
    """An `ObjectCatalogue` maintains a complete topological graph of objects with
    at most `pardim` parametric directions.
    """

    def __init__(self, pardim):
        """Initialize a catalogue for objects of parametric dimension
        `pardim`.
        """
        self.pardim = pardim
        self.count = 0

        # Internal mapping from tuples of lower-order nodes to lists of nodes
        self.internal = OrderedDict()

        # Each catalogue has a catalogue of lower dimension
        # For points, we use a VertexDict
        if pardim > 0:
            self.lower = ObjectCatalogue(pardim - 1)
        else:
            self.lower = VertexDict()

    def lookup(self, obj, add=False, raise_on_twins=()):
        """Obtain the `NodeView` object corresponding to a given object.

        If the keyword argument `add` is true, this function may generate one
        or more new nodes to accommodate the object.

        :param SplineObject obj: The object to look up
        :param bool add: Whether to allow adding new objects
        :param bool raise_on_twins: Should be a list of parametric
            dimensions where 'twins' are disallowed, i.e. two patches
            that share identical nodes of lower parametric dimension
            but which are different. For example, two surfaces with
            identical edges but different interior, like a
            'pillow'. If None, such cases will be considered two
            genuinely different patches. Setting this to a list of all
            possible parametric dimensions allow catching a number of
            common topological problems.
        :return: A corresponding view
        :rtype: NodeView

        .. warning:: The object *must* be a `SplineObject`, even for points.
        """
        # Pass lower-dimensional objects through to the lower levels
        if self.pardim > obj.pardim:
            return self.lower.lookup(obj, add=add, raise_on_twins=raise_on_twins)

        # Special case for points: self.lower is a mapping from array to node
        if self.pardim == 0:
            cps = obj.controlpoints
            if obj.rational:
                cps = cps[..., :-1]
            if add:
                node = TopologicalNode(obj, [], index=self.count)
                self.count += 1
                return self.lower.setdefault(cps, node).view()
            return self.lower[cps].view()

        # Get all nodes of lower dimension (points, vertices, etc.)
        # This involves a recursive call to self.lower.__call__
        lower_nodes = []
        for i in range(0, self.pardim):
            nodes = tuple(self.lower.lookup(obj.section(*args, unwrap_points=False), add=add,
                                            raise_on_twins=raise_on_twins).node
                          for args in sections(self.pardim, i))
            lower_nodes.append(nodes)

        # Try looking up the lower-order nodes in the internal dictionary,
        # which maps tuples of nodes to lists of nodes. E.g. for volumes we
        # look up faces, for faces we look up edges, etc. Return the first one
        # we find for which the .view() function succeeds. This can throw a
        # KeyError (if this particular combination of lower-order nodes is new)
        # or an OrientationError (if it is not new, but the objects don't
        # match). If that happens, we generate a new node and return the
        # identity view on it.
        candidates = self.internal.get(lower_nodes[-1], [])

        # If there are no candidates, we fail (unless wanting to add)
        if not candidates:
            if not add:
                raise KeyError("No such object found")
            else:
                return self._add(obj, lower_nodes)

        # If there is exactly one candidate, check it
        if len(candidates) == 1:
            try:
                return candidates[0].view(obj)
            except OrientationError:
                if self.pardim in raise_on_twins:
                    raise OrientationError(
                        "Candidate nodes found but no orientation matched. "
                        "This probably indicates an erroneous topology. "
                        "If you are sure this is not the case (that twin patches exist), "
                        "use raise_on_twins=False.",
                        candidates[0].super_owner.index,
                    )
            if not add:
                raise KeyError("No such object found")
            return self._add(obj, lower_nodes)

        # If there are multiple candidates, twins must be allowed
        if self.pardim in raise_on_twins:
            raise TwinError()
        for candidate in candidates:
            try:
                return candidate.view(obj)
            except OrientationError:
                pass
        if not add:
            raise KeyError("No such object found")
        return self._add(obj, lower_nodes)

    def add(self, obj, raise_on_twins=()):
        """Add new nodes to the graph to accommodate the given object, then return the
        corresponding `NodeView` object.

        This is equivalent to calling
        :func:`splipy.SplineModel.ObjectCatalogue.lookup` with `add` set to
        true.

        :param SplineObject obj: The object to add
        :param bool raise_on_twins: Should be a list of parametric
            dimensions where 'twins' are disallowed, i.e. two patches
            that share identical nodes of lower parametric dimension
            but which are different. For example, two surfaces with
            identical edges but different interior, like a
            'pillow'. If None, such cases will be considered two
            genuinely different patches. Setting this to a list of all
            possible parametric dimensions allow catching a number of
            common topological problems.
        :return: A corresponding view
        :rtype: NodeView

        .. warning:: The object *must* be a `SplineObject`, even for points.
        """
        return self.lookup(obj, add=True, raise_on_twins=raise_on_twins)

    def _add(self, obj, lower_nodes):
        node = TopologicalNode(obj, lower_nodes, index=self.count)
        self.count += 1
        # Assign the new node to each possible permutation of lower-order
        # nodes. This is slight overkill since some of these permutations
        # are invalid, but c'est la vie.
        for p in permutations(lower_nodes[-1]):
            self.internal.setdefault(p, []).append(node)
        return node.view()

    __call__ = add
    __getitem__ = lookup

    def top_nodes(self):
        """Return all nodes of the highest parametric dimension."""
        return self.nodes(self.pardim)

    def nodes(self, pardim):
        """Return all nodes of a given parametric dimension."""
        if self.pardim == pardim:
            if self.pardim > 0:
                return list(uniquify(chain.from_iterable(self.internal.values())))
            return list(uniquify(self.lower.values()))
        return self.lower.nodes(pardim)


# FIXME: This class is unfinished, and right now it doesn't do much other than
# wrap ObjectCatalogue

class SplineModel(object):

    def __init__(self, pardim=3, dimension=3, objs=[], force_right_hand=False):
        self.pardim = pardim
        self.dimension = dimension

        self.force_right_hand = force_right_hand
        if force_right_hand and (pardim, dimension) not in ((2,2), (3,3)):
            raise ValueError("Right-handedness only defined for 2D or 3D patches in 2D or 3D space, respectively")

        self.catalogue = ObjectCatalogue(pardim)
        self.names = {}
        self.add(objs)

    def add(self, obj, name=None, raise_on_twins=True):
        if raise_on_twins is True:
            raise_on_twins = tuple(range(self.pardim + 1))
        elif raise_on_twins is False:
            raise_on_twins = ()
        if isinstance(obj, SplineObject):
            obj = [obj]
        self._validate(obj)
        self._generate(obj, raise_on_twins=raise_on_twins)
        if name and isinstance(obj, SplineObject):
            self.names[name] = obj

    def __getitem__(self, obj):
        return self.catalogue[obj]

    def boundary(self, name=None):
        for node in self.catalogue.nodes(self.pardim-1):
            if node.nhigher == 1 and (name is None or name == node.name):
                yield node

    def assign_boundary(self, name):
        """Give a name to all unnamed boundary nodes."""
        for node in self.boundary():
            if node.name is None:
                node.name = name

    def _validate(self, objs):
        if any(p.dimension != self.dimension for p in objs):
            raise ValueError("Patches with different dimension added")
        if any(p.pardim != self.pardim for p in objs):
            raise ValueError("Patches with different parametric dimension added")
        if self.force_right_hand:
            left_inds = [i for i, p in enumerate(objs) if not is_right_hand(p)]
            if left_inds:
                indices = ', '.join(map(str, left_inds))
                raise ValueError(f"Possibly left-handed patches detected, indexes {indices}")

    def _generate(self, objs, **kwargs):
        for i, p in enumerate(objs):
            try:
                self.catalogue.add(p, **kwargs)
            except OrientationError as err:
                # TODO: Mutating exceptions is fishy.
                if len(err.args) > 1:
                    err.args = (
                        err.args[0] +
                        f" This happened while trying to connect patches at indexes"
                        f" {err.args[1]} and {i}.",
                    )
                raise err

    def generate_cp_numbers(self):
        index = 0
        for node in self.catalogue.top_nodes():
            index = node.generate_cp_numbers(index)
        self.ncps = index
        for node in self.catalogue.top_nodes():
            node.read_cp_numbers()

    def generate_cell_numbers(self):
        index = 0
        for node in self.catalogue.top_nodes():
            index = node.generate_cell_numbers(index)
        self.ncells = index

    def cps(self):
        cps = np.zeros((self.ncps, self.dimension))
        for node in self.catalogue.top_nodes():
            indices = node.cp_numbers.reshape(-1)
            values = node.obj.controlpoints.reshape(-1, self.dimension)
            cps[indices] = values
        return cps

    def faces(self):
        assert self.pardim == 3
        faces = list(chain.from_iterable(node.faces() for node in self.catalogue.top_nodes()))
        return np.hstack(faces)

    def summary(self):
        c = self.catalogue
        while isinstance(c, ObjectCatalogue):
            print('Dim {}: {}'.format(c.pardim, len(c.top_nodes())))
            c = c.lower

    def write_ifem(self, filename):
        IFEMWriter(self).write(filename)



IFEMConnection = namedtuple('IFEMConnection', ['master', 'slave', 'midx', 'sidx', 'orient'])


class IFEMWriter:

    def __init__(self, model):
        self.model = model

        # List the nodes so that the order is deterministic
        self.nodes = list(model.catalogue.top_nodes())
        self.node_ids = {node: i for i, node in enumerate(self.nodes)}

    def connections(self):
        p = self.model.pardim

        # For every object in the model...
        for node in self.nodes:

            # Loop over its sections of one lower parametric dimension
            # That is, for faces, loop over edges, and for volumes, loop over faces
            for node_sub_idx, sub in enumerate(node.lower_nodes[p - 1]):

                # Iterate over the neighbour nodes
                for neigh in set(sub.higher_nodes[p]):

                    # Only output if the node has a lower ID than the neighbour,
                    # otherwise we'll get this pair when the reverse pair is found
                    if self.node_ids[node] > self.node_ids[neigh]:
                        continue

                    # Find the section index(es) to the same sub-node as seen
                    # from the neighboring node
                    neigh_sub_idxs = [i for i, n in enumerate(neigh.lower_nodes[p - 1]) if n is sub]

                    # If the neighboring node is the same as this node (that is,
                    # the node connects to itself), we use the same rule to prevent
                    # double connections
                    if neigh is node:
                        neigh_sub_idxs = [i for i in neigh_sub_idxs if i > node_sub_idx]

                    for neigh_sub_idx in neigh_sub_idxs:
                        # Find the actual SplineObjects representing the
                        # sub-node, as it is viewed from the perspective of
                        # both the node and the neighbour, then compute the
                        # orientation mapping between them
                        node_sec_idx = section_from_index(p, p - 1, node_sub_idx)
                        node_sub = node.obj.section(*node_sec_idx, unwrap_points=False)
                        neigh_sec_idx = section_from_index(p, p - 1, neigh_sub_idx)
                        neigh_sub = neigh.obj.section(*neigh_sec_idx, unwrap_points=False)

                        orientation = Orientation.compute(node_sub, neigh_sub)

                        yield IFEMConnection(
                            master = self.node_ids[node] + 1,
                            slave = self.node_ids[neigh] + 1,
                            midx = node_sub_idx + 1,
                            sidx = neigh_sub_idx + 1,
                            orient = orientation.ifem_format,
                        )

    def write(self, filename):
        lines = [
            "<?xml version='1.0' encoding='utf-8' standalone='no'?>",
            "<topology>",
        ]

        for connection in self.connections():
                lines.append('  <connection master="{}" slave="{}" midx="{}" sidx="{}" orient="{}"/>'.format(
                    connection.master,
                    connection.slave,
                    connection.midx,
                    connection.sidx,
                    connection.orient,
                ))

        lines.extend(["</topology>"])

        with open(filename + '-topology.xinp', 'wb') as f:
            f.write('\n'.join(lines).encode('utf-8') + b'\n')

        lines = [
            "<?xml version='1.0' encoding='utf-8' standalone='no'?>",
            "<topologysets>",
        ]

        names = sorted({
            node.name for node in self.model.catalogue.nodes(self.model.pardim - 1)
            if node.name is not None
        })

        for name in names:
            entries = {}
            for node in self.model.catalogue.nodes(self.model.pardim - 1):
                if node.name != name:
                    continue
                parent = node.owner
                sub_idx = next(idx for idx, sub in enumerate(parent.lower_nodes[self.model.pardim - 1]) if sub is node)
                entries.setdefault(self.node_ids[parent], set()).add(sub_idx)
            if entries:
                kind = {2: 'face', 1: 'edge', 0: 'vertex'}[self.model.pardim - 1]
                lines.append('  <set name="{}" type="{}">'.format(name, kind))
                for node_id, sub_ids in entries.items():
                    lines.append('    <item patch="{}">{}</item>'.format(
                        node_id + 1,
                        ' '.join(str(i+1) for i in sorted(sub_ids))
                    ))
                lines.append('  </set>')

        lines.extend(["</topologysets>"])

        with open(filename + '-topologysets.xinp', 'wb') as f:
            f.write('\n'.join(lines).encode('utf-8') + b'\n')

        # Import here to avoid circular dependencies
        from .io import G2
        with G2(filename + '.g2') as f:
            f.write([n.obj for n in self.nodes])
